function [p_low,p_upp,e_low,e_upp,p_bar_11,p_bar_12,p_bar_13,p_bar_20,e_bar_11,e_bar_12,e_bar_13,e_bar_20,e_QMC]= MAXNSGP2(S,n1,n2,m1,m2,m,u,f,r,T1,T2)


%
%  Computes approximate upper and lower bounds for the maximum of a non centered non stationary Gaussian process X on rectangle, 
%  P( M_T > u ) where M_T := sup_{(s,t) in [0,T1]x[0,T2]}  (X(s,t)).
%
%  Inputs: 
%   S = number of MC replications in the  MCQMC integration method. Typical value : 12;
%   n1 = number of discretization points for the x axis for the fat grid;
%   n2 = number of discretization points for the y axis for the fat grid;
%   m1 = number of discretization points for the x axis for the thin grid;
%   m2 = number of discretization points for the y axis for the thin grid;
%   m = number of QMC  points in MCQMC integration; Typical value 1000-10000;
%   u = bound level; 
%   f = f(s,t) = E(X(s,t)) the mean function for the process vectorized; 
%   r = r(s,t,u,v) = Cov(X(s,t),X(u,v)) the covariance function for the process vectorized;
%   T1 = first  length  of the rectangle  [0,T1]x[0,T2]    ;
%   T2 = second length  of the rectangle  [0,T1]x[0,T2]    ;
%
%  Outputs: 
%   p_low = approximate lower bound, with
%   e_low = error estimate (MCQMC);
%   p_upp = approximate upper bound, with
%   e_upp = estimate for total error; 
%
%  Example:
%  [p_low,p_upp,e_low,e_upp] = MAXNSGP2(12,10,10,2,2,2000,2,@(t,s) 0,@(a,b,c,d) exp(-((a-c).^2 + (b-d).^2)/2),0,5);
%
%
%          Stephane Mourareau is the author of this program. MAXNSGP is based on Alan Genz's program MGP.
%          Stephane Mourareau, Institut Math�matiques de Toulouse, Universit� Paul Sabatier, 31062 Toulouse Cedex 9
%          Email: stephane.mourareau@math.univ-toulouse.fr
%          
% 

h1 = T1/n1;
h2 = T2/n2;

[Exp,Exp01,Exp10,Exp20,Var,Var10,Var01,Var20,Var1010,Var0101,Var1001,Var1020,Var0120,Var2020]=Cholesky(f,r,n1,n2,T1,T2); % Computation covariance matrix and expectation
[p_low,e_low]=lower_bound(S,r,n1,n2,T1,T2,m,u); % Lower bound and associated error
[pv11,ev11,p_S01]=p_11(S,m,u,n1,Var,Var10,Var1010,Exp,Exp10); %Computation sup on [0,T1] x {0} and associated error
[pv12,ev12]=p_12(S,m,u,n2,Var,Var01,Var0101,Exp,Exp01); % Computation sup on {T1} x [0,T2] and associated error
[pv13,ev13,p_S03]=p_13(S,m,u,n2,Var,Var01,Var0101,Exp,Exp01); % Computation sup on {0} x [0,T2] and associated error
[pv2,ev2]=p_22(S,m,u,n1,n2,m1,m2,Var,Var10,Var01,Var20,Var1010,Var1001,Var1020,Var0101,Var0120,Var2020,Exp,Exp01,Exp10,Exp20); % Computation inside the rectangle
po = 1 - Phi2((u-f(0,0))/sqrt(r(0,0,0,0)));
p_bar_112 = h1*( p_S01 + 2*sum(pv11(2:2:n1-2)) + pv11(n1) );
p_bar_11 = h1*( p_S01/2 + sum(pv11(1:n1-1)) + pv11(n1)/2 ); % Proba sup on [0,T1] x {0}
e_bar_11 = abs( p_bar_112 - p_bar_11)/3; % Error on [0,T1] x {0} 

p_bar_122 = h2*( pv12(1) + 2*sum(pv12(3:2:n2-1)) + pv12(n2+1) ); % Proba sup on {T1} x [0,T2]
p_bar_12 = h2*( pv12(1)/2 + sum(pv12(2:n2)) + pv12(n2+1)/2 ); % Error on {T1} x [0,T2]
e_bar_12 = abs( p_bar_122 - p_bar_12)/3;

p_bar_132 = h2*( p_S03 + 2*sum(pv13(2:2:n2-2)) + pv13(n2) ); % Proba sup on {0} x [0,T2]
p_bar_13 = h2*( p_S03/2 + sum(pv13(1:n2-1)) + pv13(n2)/2 ); % Error on {0} x [0,T2]
e_bar_13 = abs( p_bar_132 - p_bar_13)/3;

p_bar_2 = zeros(1,n1+1); 
p_bar_22 = zeros(1,n1+1); 
ev3 = zeros(1,n1*n2);
p_bar_2(1) = h2*(pv2(1)/2 + sum(pv2(2:n2)) + pv2(n2+1)/2);
p_bar_2(n1+1) = h2*(pv2(n1*(n2+1) + 1)/2 + sum(pv2((n1*(n2+1) + 2):((n1+1)*(n2+1) - 1))) + pv2((n1+1)*(n2+1))/2);
p_bar_22(1) = h2*(pv2(1) + 2*sum(pv2(3:2:n2-1)) + pv2(n2+1));
p_bar_22(n1+1) = h2*(pv2(n1*(n2+1) + 1) + 2*sum(pv2((n1*(n2+1) + 3):2:((n1+1)*(n2+1) - 2))) + pv2((n1+1)*(n2+1))/2);
for i = 1:n1-1,
    p_bar_2(i+1) = h2*(pv2(i*(n2+1) + 1)/2 + sum(pv2((i*(n2+1) + 2):(i*(n2+1) + n2))) + pv2((i+1)*(n2+1))/2);
    p_bar_22(i+1) = h2*(pv2(i*(n2+1) + 1) + 2*sum(pv2((i*(n2+1) + 3):2:(i*(n2+1) + n2 - 1))) +  pv2((i+1)*(n2+1)));
end
p_bar_20 = h1*(p_bar_2(1)/2 + sum(p_bar_2(2:n1)) + p_bar_2(n1+1)/2); % Proba sup inside [0,T1]x[0,T2]
p_bar_21 = h1*(p_bar_22(1) + 2*sum(p_bar_22(3:2:n1-1)) + p_bar_22(n1+1));
e_bar_20 = abs(p_bar_21 - p_bar_20)/3; % Corresponding error
e_QMC = sqrt( h1^2*(ev11(n1)^2 + sum(ev11(1:n1-1).^2)) + h2^2*(1/4*(ev12(1)^2 + ev12(n2+1)^2 + ev13(n2)^2) + sum(ev12(2:n2).^2) + sum(ev13(1:n2-1).^2)) + h1^2*h2^2*(1/16*(ev2(1)+ev2(n1+1)+ev2((n1+1)*(n2+1)) + ev2(n2*(n1+1)+1)) + 1/4*(sum(ev2(2:n1)) + sum(ev2((n2*(n1+1)+2):((n2+1)*(n1+1) - 1)).^2) + sum(ev2(2*(n1+1):n1+1:(n2*(n1+1))).^2) + sum(ev2(n1+2:n1+1:n2*(n1+1)+1)).^2)) + sum(sum(ev2.^2))); % Error from MCQMC


e_upp = e_bar_11 + e_bar_12 + e_bar_13 + e_bar_20 + e_QMC; %Total error
p_upp = po + p_bar_11 + p_bar_12 + p_bar_13 + p_bar_20; % Upper bound
end
